# All Lights puzzle
# written by Aleksander Sadikov, October 2007
# for AI practical classes (search topics)

class Position:
    def __init__(self, Size = 3):
        self.PuzzleSize = Size
        self.State = [-1] * self.PuzzleSize ** 2

    def ID(self):
        # ID is decimal representation of binary state (1 = 1, -1 = 0)
        return sum([v * 2 ** i for i,v in enumerate(self.State) if v > 0])

    def PrintPosition(self):
        for r in range(self.PuzzleSize):
            for c in range(self.PuzzleSize):
                print self.State[r * self.PuzzleSize + c], "\t",
            print

    def ExecuteSequence(self, sequence):
        for m in sequence:
            self.Move(m)

    def UndoSequence(self, sequence):
        for m in reversed(sequence):
            self.UndoMove(m)
    
    def Move(self, (move, cost)):
        # clicking on a square switches it and all Manhattan-adjacent squares
        r, c = divmod(move, self.PuzzleSize)
        for nr, nc in [(r, c), (r + 1, c), (r - 1, c), (r, c + 1), (r, c - 1)]:
            if nr >= 0 and nr < self.PuzzleSize and nc >= 0 and nc < self.PuzzleSize:
                self.State[nr * self.PuzzleSize + nc] *= -1

    def UndoMove(self, move):
        # undoing a move is the same as doing the same move again (-1 * -1 = 1)
        self.Move(move)

    def GenerateMoves(self):
        # clicking (switching) on any square is always possible
        # costs of all possible moves are 1 (let's say 1 click)
        return [(x,1) for x in range(self.PuzzleSize ** 2)]

    def Solved(self):
        return sum(self.State) == self.PuzzleSize ** 2

    def Evaluate(self):
        # score can simply be the number of lights that are switched on
        # negative since lower values are better (search convention!)
        return -sum(self.State) / 5.0
    
    
